#region usings
using System;
using System.Linq;
using System.ComponentModel.Composition;
using System.Reflection;
using System.Reflection.Emit;
using System.Collections.Generic;
//using System.Dynamic;

using VVVV.PluginInterfaces.V2;
using VVVV.Core.Logging;
#endregion usings

namespace VVVV.Nodes
{
	#region PluginInfo
	[PluginInfo(Name = "TypeCheck", Category = "Object", Help = "Basic template with one value in/out", Tags = "")]
	#endregion PluginInfo
	public class ObjectTypeCheckNode : IPluginEvaluate
	{
		#region fields & pins
		[Input("Input")]
		public ISpread<object> FInput;

		[Input("Learn", IsSingle = true, IsBang = true)]
		public ISpread<bool> FLearn;

		[Output("Name")]
		public ISpread<string> FName;

		[Output("Value")]
		public ISpread<ISpread<string>> FValue;

		[Import()]
		public ILogger FLogger;
		
		protected List<Func<object, object>> Getters = new List<Func<object, object>>();
		#endregion fields & pins

		//called when data for any output pin is requested
		public void Evaluate(int SpreadMax)
		{
			var FieldCount = Getters.Count;
			if (FLearn[0] && FInput.SliceCount > 0 && FInput[0] != null) {
				FLogger.Log(LogType.Debug, "hi tty!");

				var type = FInput[0].GetType();
				FieldInfo[] fields = type.GetFields();
			
				FieldCount = fields.Length;
				FName.SliceCount = FieldCount;
			
				Getters.Clear();
				for(var i=0; i< FieldCount; i++) {
					var name = fields[i].Name;
					FName[i] = name;

					var fieldInfo = type.GetField(name);
					var getter = fieldInfo.CompileGetter();
					Getters.Add(getter);
				}
			} 

			SpreadMax = FInput.SliceCount;
			FValue.SliceCount = SpreadMax;
			
			for (int i=0; i<SpreadMax;i++) {
				var fieldValues = from getter in Getters
								select getter(FInput[i]).ToString();
				
				FValue[i].AssignFrom(fieldValues);
			}
			
		}
	}
	
	public static class ReflectionUtility
	{
    	public static Func<object, object> CompileGetter(this FieldInfo field)
	    {
    	    string methodName = field.ReflectedType.FullName + ".get_" + field.Name;
        	DynamicMethod setterMethod = new DynamicMethod(methodName, typeof(object), new[] { typeof(object) }, true);
	        ILGenerator gen = setterMethod.GetILGenerator();
        	if (field.IsStatic)
    	    {
        	    gen.Emit(OpCodes.Ldsfld, field);
            	gen.Emit(field.FieldType.IsClass ? OpCodes.Castclass : OpCodes.Box, field.FieldType);
	        }
    	    else
        	{
            	gen.Emit(OpCodes.Ldarg_0);
	            gen.Emit(OpCodes.Castclass, field.DeclaringType);
    	        gen.Emit(OpCodes.Ldfld, field);
        	    gen.Emit(field.FieldType.IsClass ? OpCodes.Castclass : OpCodes.Box, field.FieldType);
	        }
    	    gen.Emit(OpCodes.Ret);
        	return (Func<object, object>)setterMethod.CreateDelegate(typeof(Func<object, object>));
    	}

    	public static Action<object, object> CompileSetter(this FieldInfo field)
    	{
        	string methodName = field.ReflectedType.FullName + ".set_" + field.Name;
        	DynamicMethod setterMethod = new DynamicMethod(methodName, null, new[] { typeof(object), typeof(object) }, true);
        	ILGenerator gen = setterMethod.GetILGenerator();
        	if (field.IsStatic)
        	{
            	gen.Emit(OpCodes.Ldarg_1);
	            gen.Emit(field.FieldType.IsClass ? OpCodes.Castclass : OpCodes.Unbox_Any, field.FieldType);
    	        gen.Emit(OpCodes.Stsfld, field);
        	}
	        else
    	    {
        	    gen.Emit(OpCodes.Ldarg_0);
            	gen.Emit(OpCodes.Castclass, field.DeclaringType);
	            gen.Emit(OpCodes.Ldarg_1);
    	        gen.Emit(field.FieldType.IsClass ? OpCodes.Castclass : OpCodes.Unbox_Any, field.FieldType);
        	    gen.Emit(OpCodes.Stfld, field);
        	}
        	gen.Emit(OpCodes.Ret);
        	return (Action<object, object>)setterMethod.CreateDelegate(typeof(Action<object, object>));
	    }
	}
}
